"""
***************************************************************************
    fillnodata.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os

from qgis.core import (
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingParameterBand,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
    QgsRasterFileWriter,
)

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils
from processing.tools.system import isWindows

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class fillnodata(GdalAlgorithm):
    INPUT = "INPUT"
    BAND = "BAND"
    DISTANCE = "DISTANCE"
    ITERATIONS = "ITERATIONS"
    NO_MASK = "NO_MASK"
    MASK_LAYER = "MASK_LAYER"
    OPTIONS = "OPTIONS"
    CREATION_OPTIONS = "CREATION_OPTIONS"
    EXTRA = "EXTRA"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterBand(
                self.BAND,
                self.tr("Band number"),
                1,
                parentLayerParameterName=self.INPUT,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.DISTANCE,
                self.tr(
                    "Maximum distance (in pixels) to search out for values to interpolate"
                ),
                type=QgsProcessingParameterNumber.Type.Integer,
                minValue=0,
                defaultValue=10,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.ITERATIONS,
                self.tr(
                    "Number of smoothing iterations to run after the interpolation"
                ),
                type=QgsProcessingParameterNumber.Type.Integer,
                minValue=0,
                defaultValue=0,
            )
        )

        # The -nomask option is no longer supported since GDAL 3.4 and
        # it doesn't work as expected even using GDAL < 3.4 https://github.com/OSGeo/gdal/pull/4201
        nomask_param = QgsProcessingParameterBoolean(
            self.NO_MASK,
            self.tr("Do not use the default validity mask for the input band"),
            defaultValue=False,
            optional=True,
        )
        nomask_param.setFlags(
            nomask_param.flags() | QgsProcessingParameterDefinition.Flag.FlagHidden
        )
        self.addParameter(nomask_param)

        self.addParameter(
            QgsProcessingParameterRasterLayer(
                self.MASK_LAYER, self.tr("Validity mask"), optional=True
            )
        )

        # backwards compatibility parameter
        # TODO QGIS 5: remove parameter and related logic
        options_param = QgsProcessingParameterString(
            self.OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        options_param.setFlags(
            options_param.flags() | QgsProcessingParameterDefinition.Flag.Hidden
        )
        options_param.setMetadata({"widget_wrapper": {"widget_type": "rasteroptions"}})
        self.addParameter(options_param)

        creation_options_param = QgsProcessingParameterString(
            self.CREATION_OPTIONS,
            self.tr("Additional creation options"),
            defaultValue="",
            optional=True,
        )
        creation_options_param.setFlags(
            creation_options_param.flags()
            | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        creation_options_param.setMetadata(
            {"widget_wrapper": {"widget_type": "rasteroptions"}}
        )
        self.addParameter(creation_options_param)

        extra_param = QgsProcessingParameterString(
            self.EXTRA,
            self.tr("Additional command-line parameters"),
            defaultValue=None,
            optional=True,
        )
        extra_param.setFlags(
            extra_param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
        )
        self.addParameter(extra_param)

        self.addParameter(
            QgsProcessingParameterRasterDestination(self.OUTPUT, self.tr("Filled"))
        )

    def name(self):
        return "fillnodata"

    def displayName(self):
        return self.tr("Fill NoData")

    def group(self):
        return self.tr("Raster analysis")

    def groupId(self):
        return "rasteranalysis"

    def commandName(self):
        return "gdal_fillnodata"

    def flags(self):
        return super().flags() | QgsProcessingAlgorithm.Flag.FlagDisplayNameIsLiteral

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        raster = self.parameterAsRasterLayer(parameters, self.INPUT, context)
        if raster is None:
            raise QgsProcessingException(
                self.invalidRasterError(parameters, self.INPUT)
            )
        input_details = GdalUtils.gdal_connection_details_from_layer(raster)

        out = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        self.setOutputValue(self.OUTPUT, out)

        arguments = [
            input_details.connection_string,
            out,
            "-md",
            str(self.parameterAsInt(parameters, self.DISTANCE, context)),
        ]

        nIterations = self.parameterAsInt(parameters, self.ITERATIONS, context)
        if nIterations:
            arguments.append("-si")
            arguments.append(str(nIterations))

        arguments.append("-b")
        arguments.append(str(self.parameterAsInt(parameters, self.BAND, context)))

        mask = self.parameterAsRasterLayer(parameters, self.MASK_LAYER, context)
        if mask:
            arguments.append("-mask")
            arguments.append(mask.source())

        output_format = self.outputFormat(parameters, self.OUTPUT, context)
        if not output_format:
            raise QgsProcessingException(self.tr("Output format is invalid"))

        arguments.append("-of")
        arguments.append(output_format)

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        if self.EXTRA in parameters and parameters[self.EXTRA] not in (None, ""):
            extra = self.parameterAsString(parameters, self.EXTRA, context)
            arguments.append(extra)

        # Until https://github.com/OSGeo/gdal/issues/7651 is fixed, creation options should be latest argument
        options = self.parameterAsString(parameters, self.CREATION_OPTIONS, context)
        # handle backwards compatibility parameter OPTIONS
        if self.OPTIONS in parameters and parameters[self.OPTIONS] not in (None, ""):
            options = self.parameterAsString(parameters, self.OPTIONS, context)
        if options:
            arguments.extend(GdalUtils.parseCreationOptions(options))

        return [
            self.commandName() + (".bat" if isWindows() else ".py"),
            GdalUtils.escapeAndJoin(arguments),
        ]



"""
/***************************************************************************
                            Plugin Installer module
                             -------------------
    Date                 : May 2013
    Copyright            : (C) 2013 by Borys Jurgiel
    Email                : info at borysjurgiel dot pl

    This module is based on former plugin_installer plugin:
      Copyright (C) 2007-2008 Matthew Perry
      Copyright (C) 2008-2013 Borys Jurgiel

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import json
import os
import shutil
import zipfile

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsNetworkAccessManager,
    QgsNetworkRequestParameters,
    QgsSettings,
    QgsSettingsTree,
)
from qgis.gui import QgsHelp, QgsPasswordLineEdit
from qgis.PyQt import sip
from qgis.PyQt.QtCore import (
    QDateTime,
    QDir,
    QFileInfo,
    QObject,
    Qt,
    QTemporaryDir,
    QUrl,
)
from qgis.PyQt.QtNetwork import QNetworkRequest
from qgis.PyQt.QtWidgets import (
    QApplication,
    QDialog,
    QDialogButtonBox,
    QFrame,
    QLabel,
    QMessageBox,
    QVBoxLayout,
)
from qgis.utils import (
    HOME_PLUGIN_PATH,
    OverrideCursor,
    iface,
    isPluginLoaded,
    loadPlugin,
    plugins_metadata_parser,
    startPlugin,
    unloadPlugin,
    updateAvailablePlugins,
)

from .installer_data import officialRepo, plugins, removeDir, reposGroup, repositories
from .plugindependencies import find_dependencies
from .qgsplugindependenciesdialog import QgsPluginDependenciesDialog
from .qgsplugininstallerfetchingdialog import QgsPluginInstallerFetchingDialog
from .qgsplugininstallerinstallingdialog import QgsPluginInstallerInstallingDialog
from .qgsplugininstallerpluginerrordialog import QgsPluginInstallerPluginErrorDialog
from .qgsplugininstallerrepositorydialog import QgsPluginInstallerRepositoryDialog
from .unzip import unzip

# public instances:
pluginInstaller = None


def initPluginInstaller():
    global pluginInstaller
    pluginInstaller = QgsPluginInstaller()


# -------------------------------------------------------- #
class QgsPluginInstaller(QObject):
    """The main class for managing the plugin installer stuff"""

    # ----------------------------------------- #
    def __init__(self):
        """Initialize data objects, starts fetching if appropriate, and warn about/removes obsolete plugins"""

        QObject.__init__(self)  # initialize QObject in order to to use self.tr()
        repositories.load()
        plugins.getAllInstalled()

        self.message_bar_widget = None

        if (
            repositories.checkingOnStart()
            and repositories.timeForChecking()
            and repositories.allEnabled()
        ):
            # start fetching repositories
            repositories.checkingDone.connect(self.checkingDone)
            for key in repositories.allEnabled():
                repositories.requestFetching(key)
        else:
            # no fetching at start, so mark all enabled repositories as requesting to be fetched.
            for key in repositories.allEnabled():
                repositories.setRepositoryData(key, "state", 3)

        # look for obsolete plugins updates (the user-installed one is older than the core one)
        for key in plugins.obsoletePlugins:
            plugin = plugins.localCache[key]
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Icon.Warning)
            msg.setWindowTitle(self.tr("QGIS Python Plugin Installer"))
            msg.addButton(
                self.tr("Uninstall (recommended)"), QMessageBox.ButtonRole.AcceptRole
            )
            msg.addButton(
                self.tr("I will uninstall it later"), QMessageBox.ButtonRole.RejectRole
            )
            msg.setText(
                "{} <b>{}</b><br/><br/>{}".format(
                    self.tr("Obsolete plugin:"),
                    plugin["name"],
                    self.tr(
                        "QGIS has detected an obsolete plugin that masks its more recent version shipped with this copy of QGIS. This is likely due to files associated with a previous installation of QGIS. Do you want to remove the old plugin right now and unmask the more recent version?"
                    ),
                )
            )
            msg.exec()
            if not msg.result():
                settings = QgsSettings()
                plugin_is_active = settings.value(
                    "/PythonPlugins/" + key, False, type=bool
                )

                # uninstall the update, update utils and reload if enabled
                self.uninstallPlugin(key, quiet=True)
                updateAvailablePlugins()
                if plugin_is_active:
                    settings.setValue(
                        "/PythonPlugins/watchDogTimestamp/" + key,
                        QDateTime.currentDateTime().toSecsSinceEpoch(),
                    )
                    loadPlugin(key)
                    startPlugin(key)
                    settings.remove("/PythonPlugins/watchDogTimestamp/" + key)

    # ----------------------------------------- #
    def fetchAvailablePlugins(self, reloadMode):
        """Fetch plugins from all enabled repositories."""
        """  reloadMode = true:  Fully refresh data from QgsSettings to mRepositories  """
        """  reloadMode = false: Fetch unready repositories only """
        if reloadMode:
            repositories.load()
            plugins.clearRepoCache()
            plugins.getAllInstalled()

        for key in repositories.allEnabled():
            if (
                reloadMode or repositories.all()[key]["state"] == 3
            ):  # if state = 3 (error or not fetched yet), try to fetch once again
                repositories.requestFetching(key, force_reload=reloadMode)

        if repositories.fetchingInProgress():
            fetchDlg = QgsPluginInstallerFetchingDialog(iface.mainWindow())
            fetchDlg.exec()
            del fetchDlg
            for key in repositories.all():
                repositories.killConnection(key)

        # display error messages for every unavailable repository, unless Shift pressed nor all repositories are unavailable
        keepQuiet = QgsApplication.keyboardModifiers() == Qt.KeyboardModifiers(
            Qt.KeyboardModifier.ShiftModifier
        )
        if (
            repositories.allUnavailable()
            and repositories.allUnavailable() != repositories.allEnabled()
        ):
            for key in repositories.allUnavailable():
                if not keepQuiet:
                    QMessageBox.warning(
                        iface.mainWindow(),
                        self.tr("QGIS Python Plugin Installer"),
                        self.tr("Error reading repository:")
                        + " "
                        + key
                        + "\n\n"
                        + repositories.all()[key]["error"],
                    )
                if QgsApplication.keyboardModifiers() == Qt.KeyboardModifiers(
                    Qt.KeyboardModifier.ShiftModifier
                ):
                    keepQuiet = True
        # finally, rebuild plugins from the caches
        plugins.rebuild()

    # ----------------------------------------- #
    def checkingDone(self):
        """Remove the "Looking for new plugins..." label and display a notification instead if any updates available"""

        # rebuild plugins cache
        plugins.rebuild()
        # look for news in the repositories
        plugins.markNews()
        # then check for updates (and eventually overwrite status)
        updatable_plugin_names = []
        for _, properties in plugins.all().items():
            if properties["status"] == "upgradeable":
                updatable_plugin_names.append(properties["name"])

        if not updatable_plugin_names:
            return

        iface.mainWindow().pluginUpdatesAvailable.emit(updatable_plugin_names)

    # ----------------------------------------- #
    def exportRepositoriesToManager(self):
        """Update manager's repository tree widget with current data"""
        iface.pluginManagerInterface().clearRepositoryList()
        for key in repositories.all():
            url = repositories.all()[key]["url"] + repositories.urlParams()
            if repositories.inspectionFilter():
                enabled = key == repositories.inspectionFilter()
            else:
                enabled = repositories.all()[key]["enabled"]
            iface.pluginManagerInterface().addToRepositoryList(
                {
                    "name": key,
                    "url": url,
                    "enabled": enabled and "true" or "false",
                    "valid": repositories.all()[key]["valid"] and "true" or "false",
                    "state": str(repositories.all()[key]["state"]),
                    "error": repositories.all()[key]["error"],
                    "inspection_filter": repositories.inspectionFilter()
                    and "true"
                    or "false",
                }
            )

    # ----------------------------------------- #
    def exportPluginsToManager(self):
        """Insert plugins metadata to QgsMetadataRegistry"""
        iface.pluginManagerInterface().clearPythonPluginMetadata()
        for key in plugins.all():
            plugin = plugins.all()[key]
            iface.pluginManagerInterface().addPluginMetadata(
                {
                    "id": key,
                    "plugin_id": plugin["plugin_id"] or "",
                    "name": plugin["name"],
                    "description": plugin["description"],
                    "about": plugin["about"],
                    "category": plugin["category"],
                    "tags": plugin["tags"],
                    "changelog": plugin["changelog"],
                    "author_name": plugin["author_name"],
                    "author_email": plugin["author_email"],
                    "homepage": plugin["homepage"],
                    "tracker": plugin["tracker"],
                    "code_repository": plugin["code_repository"],
                    "version_installed": plugin["version_installed"],
                    "library": plugin["library"],
                    "icon": plugin["icon"],
                    "readonly": plugin["readonly"] and "true" or "false",
                    "installed": plugin["installed"] and "true" or "false",
                    "available": plugin["available"] and "true" or "false",
                    "status": plugin["status"],
                    "status_exp": plugin["status_exp"],
                    "error": plugin["error"],
                    "error_details": plugin["error_details"],
                    "create_date": plugin["create_date"],
                    "update_date": plugin["update_date"],
                    "create_date_stable": plugin["create_date_stable"],
                    "update_date_stable": plugin["update_date_stable"],
                    "create_date_experimental": plugin["create_date_experimental"],
                    "update_date_experimental": plugin["update_date_experimental"],
                    "experimental": plugin["experimental"] and "true" or "false",
                    "deprecated": plugin["deprecated"] and "true" or "false",
                    "trusted": plugin["trusted"] and "true" or "false",
                    "version_available": plugin["version_available"],
                    "version_available_stable": plugin["version_available_stable"]
                    or "",
                    "version_available_experimental": plugin[
                        "version_available_experimental"
                    ]
                    or "",
                    "zip_repository": plugin["zip_repository"],
                    "download_url": plugin["download_url"],
                    "download_url_stable": plugin["download_url_stable"],
                    "download_url_experimental": plugin["download_url_experimental"],
                    "filename": plugin["filename"],
                    "downloads": plugin["downloads"],
                    "average_vote": plugin["average_vote"],
                    "rating_votes": plugin["rating_votes"],
                    "plugin_dependencies": plugin.get("plugin_dependencies", None),
                    "pythonic": "true",
                }
            )
        iface.pluginManagerInterface().reloadModel()

    # ----------------------------------------- #
    def reloadAndExportData(self):
        """Reload All repositories and export data to the Plugin Manager"""
        self.fetchAvailablePlugins(reloadMode=True)
        self.exportRepositoriesToManager()
        self.exportPluginsToManager()

    # ----------------------------------------- #
    def showPluginManagerWhenReady(
        self, tab_index: int = -1, search_term: str = ""
    ) -> None:
        """
        Open the plugin manager window.

        If fetching is still in progress, it shows the progress window first
        Optionally pass the index of tab to be opened and a search term
        to filter plugins in the manager

        :param tab_index: index of the tab to be opened
        :type tab_index: int
        :param search_term: text used to filter plugins in the manager
        :type search_term: str
        """
        if self.message_bar_widget:
            if not sip.isdeleted(self.message_bar_widget):
                iface.messageBar().popWidget(self.message_bar_widget)
            self.message_bar_widget = None

        self.fetchAvailablePlugins(reloadMode=False)
        self.exportRepositoriesToManager()
        self.exportPluginsToManager()

        # Finally, show the plugin manager window
        # Ensure that the index is within a valid range
        try:
            tab_index = int(tab_index)
            if tab_index < 0 or tab_index > 6:
                tab_index = -1
        except (ValueError, TypeError):
            tab_index = -1

        iface.pluginManagerInterface().showPluginManager(tab_index, search_term)

    # ----------------------------------------- #
    def onManagerClose(self):
        """Call this method when closing manager window - it resets last-use-dependent values."""
        plugins.updateSeenPluginsList()
        repositories.saveCheckingOnStartLastDate()

    # ----------------------------------------- #
    def exportSettingsGroup(self):
        """Return QgsSettings settingsGroup value"""
        # todo QGIS 5 remove
        return "plugin-manager"

    # ----------------------------------------- #
    def upgradeAllUpgradeable(self):
        """Reinstall all upgradeable plugins"""
        for key in plugins.allUpgradeable():
            self.installPlugin(key, quiet=True)

    # ----------------------------------------- #
    def installPlugin(self, key, quiet=False, stable=True):
        """Install given plugin"""
        error = False
        status_key = "status" if stable else "status_exp"
        infoString = ("", "")
        plugin = plugins.all()[key]
        previousStatus = plugin[status_key]
        if not plugin:
            return
        if (
            plugin[status_key] == "newer" and not plugin["error"]
        ):  # ask for confirmation if user downgrades an usable plugin
            if (
                QMessageBox.warning(
                    iface.mainWindow(),
                    self.tr("QGIS Python Plugin Installer"),
                    self.tr(
                        "Are you sure you want to downgrade the plugin to the latest available version? The installed one is newer!"
                    ),
                    QMessageBox.StandardButton.Yes,
                    QMessageBox.StandardButton.No,
                )
                == QMessageBox.StandardButton.No
            ):
                return

        # if plugin is active, unload it before update, see https://github.com/qgis/QGIS/issues/54968
        pluginWasLoaded = isPluginLoaded(plugin["id"])
        if pluginWasLoaded:
            unloadPlugin(plugin["id"])

        dlg = QgsPluginInstallerInstallingDialog(
            iface.mainWindow(), plugin, stable=stable
        )
        dlg.exec()

        plugin_path = HOME_PLUGIN_PATH + "/" + key
        if dlg.result():
            error = True
            infoString = (self.tr("Plugin installation failed"), dlg.result())
            # download failed or aborted. If plugin was active before the update, let's try to load it back
            if pluginWasLoaded and loadPlugin(plugin["id"]):
                startPlugin(plugin["id"])
        elif not QDir(plugin_path).exists():
            error = True
            infoString = (
                self.tr("Plugin has disappeared"),
                self.tr(
                    'The plugin seems to have been installed but it\'s not possible to know where. The directory "{}" '
                    "has not been found. Probably the plugin package contained a wrong named directory.\nPlease search "
                    "the list of installed plugins. You should find the plugin there, but it's not possible to "
                    "determine which of them it is and it's also not possible to inform you about available updates. "
                    "Please contact the plugin author and submit this issue."
                ).format(plugin_path),
            )
            with OverrideCursor(Qt.CursorShape.WaitCursor):
                plugins.getAllInstalled()
                plugins.rebuild()
                self.exportPluginsToManager()
        else:
            QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
            # update the list of plugins in plugin handling routines
            updateAvailablePlugins()
            self.processDependencies(plugin["id"])
            # try to load the plugin
            loadPlugin(plugin["id"])
            plugins.getAllInstalled()
            plugins.rebuild()
            plugin = plugins.all()[key]
            if not plugin["error"]:
                if previousStatus in ["not installed", "new"]:
                    infoString = (self.tr("Plugin installed successfully"), "")
                    if startPlugin(plugin["id"]):
                        settings = QgsSettings()
                        settings.setValue("/PythonPlugins/" + plugin["id"], True)
                else:
                    infoString = (self.tr("Plugin reinstalled successfully"), "")
                    if pluginWasLoaded:
                        loadPlugin(plugin["id"])
                        startPlugin(plugin["id"])
                    else:
                        unloadPlugin(
                            key
                        )  # Just for a case. Will exit quietly if really not loaded
                        loadPlugin(key)
                if quiet:
                    infoString = (None, None)
                QApplication.restoreOverrideCursor()
            else:
                QApplication.restoreOverrideCursor()
                if plugin["error"] == "incompatible":
                    message = self.tr(
                        "The plugin is not compatible with this version of QGIS. It's designed for QGIS versions:"
                    )
                    message += " <b>" + plugin["error_details"] + "</b>"
                elif plugin["error"] == "dependent":
                    message = self.tr(
                        "The plugin depends on some components missing on your system. You need to install the following Python module in order to enable it:"
                    )
                    message += "<b> " + plugin["error_details"] + "</b>"
                else:
                    message = self.tr("The plugin is broken. Python said:")
                    message += "<br><b>" + plugin["error_details"] + "</b>"
                dlg = QgsPluginInstallerPluginErrorDialog(iface.mainWindow(), message)
                dlg.exec()
                if dlg.result():
                    # revert installation
                    pluginDir = HOME_PLUGIN_PATH + "/" + plugin["id"]
                    result = removeDir(pluginDir)
                    if QDir(pluginDir).exists():
                        error = True
                        infoString = (self.tr("Plugin uninstall failed"), result)
                        try:
                            exec("sys.path_importer_cache.clear()")
                            exec("import %s" % plugin["id"])
                            exec("reload (%s)" % plugin["id"])
                        except:
                            pass
                    else:
                        try:
                            exec("del sys.modules[%s]" % plugin["id"])
                        except:
                            pass
                    plugins.getAllInstalled()
                    plugins.rebuild()

            self.exportPluginsToManager()

        if infoString[0]:
            level = error and Qgis.MessageLevel.Critical or Qgis.MessageLevel.Info
            msg = "<b>%s</b>" % infoString[0]
            if infoString[1]:
                msg += "<b>:</b> %s" % infoString[1]
            iface.pluginManagerInterface().pushMessage(msg, level)

    # ----------------------------------------- #
    def uninstallPlugin(self, key, quiet=False):
        """Uninstall given plugin"""
        if key in plugins.all():
            plugin = plugins.all()[key]
        else:
            plugin = plugins.localCache[key]
        if not plugin:
            return
        if not quiet:
            warning = (
                self.tr("Are you sure you want to uninstall the following plugin?")
                + "\n("
                + plugin["name"]
                + ")"
            )
            if (
                plugin["status"] == "orphan"
                and plugin["status_exp"] == "orphan"
                and not plugin["error"]
            ):
                warning += "\n\n" + self.tr(
                    "Warning: this plugin isn't available in any accessible repository!"
                )
            if (
                QMessageBox.warning(
                    iface.mainWindow(),
                    self.tr("QGIS Python Plugin Installer"),
                    warning,
                    QMessageBox.StandardButton.Yes,
                    QMessageBox.StandardButton.No,
                )
                == QMessageBox.StandardButton.No
            ):
                return
        # unload the plugin
        QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
        try:
            unloadPlugin(key)
        except:
            pass
        pluginDir = HOME_PLUGIN_PATH + "/" + plugin["id"]
        result = removeDir(pluginDir)
        if result:
            QApplication.restoreOverrideCursor()
            msg = "<b>{}:</b>{}".format(self.tr("Plugin uninstall failed"), result)
            iface.pluginManagerInterface().pushMessage(msg, Qgis.MessageLevel.Critical)
        else:
            # safe remove
            try:
                unloadPlugin(plugin["id"])
            except:
                pass
            try:
                exec("plugins[%s].unload()" % plugin["id"])
                exec("del plugins[%s]" % plugin["id"])
            except:
                pass
            try:
                exec("del sys.modules[%s]" % plugin["id"])
            except:
                pass
            try:
                exec("del plugins_metadata_parser[%s]" % plugin["id"])
            except:
                pass

            plugins.getAllInstalled()
            plugins.rebuild()
            self.exportPluginsToManager()
            QApplication.restoreOverrideCursor()
            iface.pluginManagerInterface().pushMessage(
                self.tr("Plugin uninstalled successfully"), Qgis.MessageLevel.Success
            )

            settings = QgsSettings()
            settings.remove("/PythonPlugins/" + key)

    # ----------------------------------------- #
    def addRepository(self):
        """add new repository connection"""
        dlg = QgsPluginInstallerRepositoryDialog(iface.mainWindow())
        dlg.editParams.setText(repositories.urlParams())
        dlg.checkBoxEnabled.setCheckState(Qt.CheckState.Checked)
        if not dlg.exec():
            return
        for i in list(repositories.all().values()):
            if dlg.editURL.text().strip() == i["url"]:
                iface.pluginManagerInterface().pushMessage(
                    self.tr("Unable to add another repository with the same URL!"),
                    Qgis.MessageLevel.Warning,
                )
                return
        settings = QgsSettings()
        settings.beginGroup(reposGroup)
        reposName = dlg.editName.text()
        reposURL = dlg.editURL.text().strip()
        if reposName in repositories.all():
            reposName = reposName + "(2)"
        # add to settings
        settings.setValue(reposName + "/url", reposURL)
        settings.setValue(reposName + "/authcfg", dlg.configId().strip())
        settings.setValue(
            reposName + "/enabled", bool(dlg.checkBoxEnabled.checkState())
        )
        # refresh lists and populate widgets
        plugins.removeRepository(reposName)
        self.reloadAndExportData()

    # ----------------------------------------- #
    def editRepository(self, reposName):
        """edit repository connection"""
        if not reposName:
            return
        checkState = {False: Qt.CheckState.Unchecked, True: Qt.CheckState.Checked}
        dlg = QgsPluginInstallerRepositoryDialog(iface.mainWindow())
        dlg.editName.setText(reposName)
        dlg.editURL.setText(repositories.all()[reposName]["url"])
        dlg.editAuthCfgWgt.setConfigId(repositories.all()[reposName]["authcfg"])
        dlg.editParams.setText(repositories.urlParams())
        dlg.checkBoxEnabled.setCheckState(
            checkState[repositories.all()[reposName]["enabled"]]
        )
        if repositories.all()[reposName]["valid"]:
            dlg.checkBoxEnabled.setEnabled(True)
            dlg.labelInfo.setText("")
        else:
            dlg.checkBoxEnabled.setEnabled(False)
            dlg.labelInfo.setText(
                self.tr(
                    "This repository is blocked due to incompatibility with your QGIS version"
                )
            )
            dlg.labelInfo.setFrameShape(QFrame.Shape.Box)
        if not dlg.exec():
            return  # nothing to do if canceled
        for i in list(repositories.all().values()):
            if (
                dlg.editURL.text().strip() == i["url"]
                and dlg.editURL.text().strip() != repositories.all()[reposName]["url"]
            ):
                iface.pluginManagerInterface().pushMessage(
                    self.tr("Unable to add another repository with the same URL!"),
                    Qgis.MessageLevel.Warning,
                )
                return
        # delete old repo from QgsSettings and create new one
        settings = QgsSettings()
        settings.beginGroup(reposGroup)
        settings.remove(reposName)
        newName = dlg.editName.text()
        if newName in repositories.all() and newName != reposName:
            newName = newName + "(2)"
        settings.setValue(newName + "/url", dlg.editURL.text().strip())
        settings.setValue(newName + "/authcfg", dlg.configId().strip())
        settings.setValue(newName + "/enabled", bool(dlg.checkBoxEnabled.checkState()))
        if (
            dlg.editAuthCfgWgt.configId().strip()
            != repositories.all()[reposName]["authcfg"]
        ):
            repositories.all()[reposName]["authcfg"] = (
                dlg.editAuthCfgWgt.configId().strip()
            )
        if (
            dlg.editURL.text().strip() == repositories.all()[reposName]["url"]
            and dlg.checkBoxEnabled.checkState()
            == checkState[repositories.all()[reposName]["enabled"]]
        ):
            repositories.rename(reposName, newName)
            self.exportRepositoriesToManager()
            return  # nothing else to do if only repository name was changed
        plugins.removeRepository(reposName)
        self.reloadAndExportData()

    # ----------------------------------------- #
    def deleteRepository(self, reposName: str):
        """delete repository connection"""
        if not reposName:
            return
        settings = QgsSettings()
        settings.beginGroup(reposGroup)
        if settings.value(reposName + "/url", "", type=str) == officialRepo[1]:
            iface.pluginManagerInterface().pushMessage(
                self.tr(
                    "You can't remove the official QGIS Plugin Repository. You can disable it if needed."
                ),
                Qgis.MessageLevel.Warning,
            )
            return
        warning = (
            self.tr("Are you sure you want to remove the following repository?")
            + "\n"
            + reposName
        )
        if (
            QMessageBox.warning(
                iface.mainWindow(),
                self.tr("QGIS Python Plugin Installer"),
                warning,
                QMessageBox.StandardButton.Yes,
                QMessageBox.StandardButton.No,
            )
            == QMessageBox.StandardButton.No
        ):
            return
        # delete from the settings, refresh data and repopulate all the widgets
        settings.remove(reposName)
        repositories.remove(reposName)
        plugins.removeRepository(reposName)
        self.reloadAndExportData()

    # ----------------------------------------- #
    def setRepositoryInspectionFilter(self, reposName=None):
        """temporarily block another repositories to fetch only one for inspection"""
        repositories.setInspectionFilter(reposName)
        self.reloadAndExportData()

    # ----------------------------------------- #
    def sendVote(self, plugin_id, vote):
        """send vote via the RPC"""

        if not plugin_id or not vote:
            return False
        url = "https://plugins.qgis.org/plugins/RPC2/"
        params = {
            "id": "djangorpc",
            "method": "plugin.vote",
            "params": [str(plugin_id), str(vote)],
        }
        req = QNetworkRequest(QUrl(url))
        req.setAttribute(
            QNetworkRequest.Attribute(
                QgsNetworkRequestParameters.RequestAttributes.AttributeInitiatorClass
            ),
            "QgsPluginInstaller",
        )
        req.setAttribute(
            QNetworkRequest.Attribute(
                QgsNetworkRequestParameters.RequestAttributes.AttributeInitiatorRequestId
            ),
            "sendVote",
        )
        req.setRawHeader(b"Content-Type", b"application/json")
        reply = QgsNetworkAccessManager.instance().blockingPost(
            req, bytes(json.dumps(params), "utf-8")
        )
        if reply.attribute(QNetworkRequest.Attribute.HttpStatusCodeAttribute) == 200:
            return True
        else:
            return False

    def installFromZipFile(self, filePath):
        if not os.path.isfile(filePath):
            return False

        QgsSettingsTree.node("plugin-manager").childSetting(
            "last-zip-directory"
        ).setValue(QFileInfo(filePath).absoluteDir().absolutePath())

        pluginName = None
        with zipfile.ZipFile(filePath, "r") as zf:
            # search for metadata.txt. In case of multiple files, we can assume that
            # the shortest path relates <pluginname>/metadata.txt
            metadatafiles = sorted(
                f for f in zf.namelist() if f.endswith("metadata.txt")
            )
            if len(metadatafiles) > 0:
                pluginName = os.path.split(metadatafiles[0])[0]

        if not pluginName:
            msg_box = QMessageBox()
            msg_box.setIcon(QMessageBox.Icon.Warning)
            msg_box.setWindowTitle(
                self.tr("QGIS Python Install from ZIP Plugin Installer")
            )
            msg_box.setText(
                self.tr(
                    "The Zip file is not a valid QGIS python plugin. No root folder was found inside."
                )
            )
            msg_box.setStandardButtons(QMessageBox.StandardButton.Ok)
            more_info_btn = msg_box.addButton(
                self.tr("More Information"), QMessageBox.ButtonRole.HelpRole
            )
            msg_box.exec()
            if msg_box.clickedButton() == more_info_btn:
                QgsHelp.openHelp("plugins/plugins.html#the-install-from-zip-tab")
            return False

        pluginsDirectory = HOME_PLUGIN_PATH
        if not QDir(pluginsDirectory).exists():
            QDir().mkpath(pluginsDirectory)

        pluginDirectory = QDir.cleanPath(os.path.join(pluginsDirectory, pluginName))

        # if plugin is active, unload it before update, see https://github.com/qgis/QGIS/issues/54968
        if isPluginLoaded(pluginName):
            unloadPlugin(pluginName)

        password = None
        infoString = None
        success = False
        keepTrying = True

        extractDir = QTemporaryDir()
        while keepTrying:
            try:
                # Test extraction. If fails, then exception will be raised and no removing occurs
                unzip(filePath, extractDir.path(), password)
                keepTrying = False
                success = True
            except Exception as e:
                success = False
                if "password" in str(e):
                    infoString = self.tr("Aborted by user")
                    if "Bad password" in str(e):
                        msg = self.tr(
                            "Wrong password. Please enter a correct password to the zip file."
                        )
                    else:
                        msg = self.tr(
                            "The zip file is encrypted. Please enter password."
                        )
                    # Display a password dialog with QgsPasswordLineEdit
                    dlg = QDialog()
                    dlg.setWindowTitle(self.tr("Enter password"))
                    buttonBox = QDialogButtonBox(
                        QDialogButtonBox.StandardButton.Ok
                        | QDialogButtonBox.StandardButton.Cancel,
                        Qt.Orientation.Horizontal,
                    )
                    buttonBox.rejected.connect(dlg.reject)
                    buttonBox.accepted.connect(dlg.accept)
                    lePass = QgsPasswordLineEdit()
                    layout = QVBoxLayout()
                    layout.addWidget(QLabel(msg))
                    layout.addWidget(lePass)
                    layout.addWidget(buttonBox)
                    dlg.setLayout(layout)
                    keepTrying = dlg.exec()
                    password = lePass.text()
                else:
                    infoString = self.tr(
                        f"Failed to unzip the plugin package\n{filePath}.\nProbably it is broken"
                    )
                    keepTrying = False

        if success:
            # Removing old plugin files if exist
            # If the target directory already exists as a link,
            # the link should be removed without resolving
            infoString = removeDir(pluginDirectory)
            if infoString:
                success = False

            try:
                shutil.move(extractDir.filePath(pluginName), pluginsDirectory)
            except:
                infoString = (
                    self.tr("Could not store plugin to the plugin directory:")
                    + "\n"
                    + pluginDirectory
                )
                success = False

        if success:
            with OverrideCursor(Qt.CursorShape.WaitCursor):
                updateAvailablePlugins()
                self.processDependencies(pluginName)
                loadPlugin(pluginName)
                plugins.getAllInstalled()
                plugins.rebuild()

                settings = QgsSettings()
                if settings.contains(
                    "/PythonPlugins/" + pluginName
                ):  # Plugin was available?
                    unloadPlugin(pluginName)
                    loadPlugin(pluginName)
                    if settings.value(
                        "/PythonPlugins/" + pluginName, False, bool
                    ):  # Plugin was also active?
                        startPlugin(pluginName)
                else:
                    if startPlugin(pluginName):
                        settings.setValue("/PythonPlugins/" + pluginName, True)

            self.exportPluginsToManager()
            msg = "<b>%s</b>" % self.tr("Plugin installed successfully")
        else:
            msg = "<b>{}:</b> {}".format(
                self.tr("Plugin installation failed"), infoString
            )

        level = Qgis.MessageLevel.Success if success else Qgis.MessageLevel.Critical
        iface.pluginManagerInterface().pushMessage(msg, level)
        return success

    def processDependencies(self, plugin_id):
        """Processes plugin dependencies

        :param plugin_id: plugin id
        :type plugin_id: str
        """

        to_install, to_upgrade, not_found = find_dependencies(plugin_id)
        if to_install or to_upgrade or not_found:
            dlg = QgsPluginDependenciesDialog(
                plugin_id, to_install, to_upgrade, not_found
            )
            if dlg.exec() == QDialog.DialogCode.Accepted:
                actions = dlg.actions()
                for dependency_plugin_id, action_data in actions.items():
                    try:
                        self.installPlugin(
                            dependency_plugin_id,
                            stable=action_data["use_stable_version"],
                        )
                        if action_data["action"] == "install":
                            iface.pluginManagerInterface().pushMessage(
                                self.tr(
                                    "Plugin dependency <b>%s</b> successfully installed"
                                )
                                % dependency_plugin_id,
                                Qgis.MessageLevel.Success,
                            )
                        else:
                            iface.pluginManagerInterface().pushMessage(
                                self.tr(
                                    "Plugin dependency <b>%s</b> successfully upgraded"
                                )
                                % dependency_plugin_id,
                                Qgis.MessageLevel.Success,
                            )
                    except Exception as ex:
                        if action_data["action"] == "install":
                            iface.pluginManagerInterface().pushMessage(
                                self.tr(
                                    "Error installing plugin dependency <b>%s</b>: %s"
                                )
                                % (dependency_plugin_id, ex),
                                Qgis.MessageLevel.Warning,
                            )
                        else:
                            iface.pluginManagerInterface().pushMessage(
                                self.tr(
                                    "Error upgrading plugin dependency <b>%s</b>: %s"
                                )
                                % (dependency_plugin_id, ex),
                                Qgis.MessageLevel.Warning,
                            )



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : Oct 14 2016
copyright            : (C) 2016 by Even Rouault
                       (C) 2011 by Giuseppe Sucameli
email                : even.rouault at spatialys.com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import sqlite3
from functools import cmp_to_key

from osgeo import gdal, ogr, osr
from qgis.core import (
    QgsAbstractDatabaseProviderConnection,
    QgsApplication,
    QgsProviderConnectionException,
    QgsProviderRegistry,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QThread
from qgis.PyQt.QtWidgets import QApplication
from qgis.utils import spatialite_connect

from ..connector import DBConnector
from ..plugin import ConnectionError, DbError, Table

gdal.UseExceptions()
ogr.UseExceptions()


def classFactory():
    return GPKGDBConnector


class GPKGDBConnector(DBConnector):
    def __init__(self, uri, connection):
        """Creates a new GPKG connector

        :param uri: data source URI
        :type uri: QgsDataSourceUri
        :param connection: the GPKGDBPlugin parent instance
        :type connection: GPKGDBPlugin
        """

        DBConnector.__init__(self, uri)
        self.dbname = uri.database()
        self.connection = connection
        self._current_thread = None
        md = QgsProviderRegistry.instance().providerMetadata(connection.providerName())
        # QgsAbstractDatabaseProviderConnection instance
        self.core_connection = md.findConnection(connection.connectionName())
        if self.core_connection is None:
            self.core_connection = md.createConnection(uri.uri(), {})
        self.has_raster = False
        self.mapSridToName = {}
        # To be removed when migration to new API is completed
        self._opendb()

    def _opendb(self):

        # Keep this explicit assignment to None to make sure the file is
        # properly closed before being re-opened
        self.gdal_ds = None
        try:
            self.gdal_ds = gdal.OpenEx(self.dbname, gdal.OF_UPDATE)
        except Exception:
            try:
                self.gdal_ds = gdal.OpenEx(self.dbname)
            except Exception:
                raise ConnectionError(
                    QApplication.translate("DBManagerPlugin", '"{0}" not found').format(
                        self.dbname
                    )
                )
        if self.gdal_ds.GetDriver().ShortName != "GPKG":
            raise ConnectionError(
                QApplication.translate(
                    "DBManagerPlugin",
                    '"{dbname}" not recognized as GPKG ({shortname} reported instead.)',
                ).format(
                    dbname=self.dbname, shortname=self.gdal_ds.GetDriver().ShortName
                )
            )
        self.has_raster = (
            self.gdal_ds.RasterCount != 0
            or self.gdal_ds.GetMetadata("SUBDATASETS") is not None
        )
        self.connection = None
        self._current_thread = None

    @property
    def connection(self):
        """Creates and returns a spatialite connection, if
        the existing connection was created in another thread
        invalidates it and create a new one.
        """

        if self._connection is None or self._current_thread != int(
            QThread.currentThreadId()
        ):
            self._current_thread = int(QThread.currentThreadId())
            try:
                self._connection = spatialite_connect(str(self.dbname))
            except self.connection_error_types() as e:
                raise ConnectionError(e)
        return self._connection

    @connection.setter
    def connection(self, conn):
        self._connection = conn

    def unquoteId(self, quotedId):
        if (
            len(quotedId) <= 2
            or quotedId[0] != '"'
            or quotedId[len(quotedId) - 1] != '"'
        ):
            return quotedId
        unquoted = ""
        i = 1
        while i < len(quotedId) - 1:
            if quotedId[i] == '"' and quotedId[i + 1] == '"':
                unquoted += '"'
                i += 2
            else:
                unquoted += quotedId[i]
                i += 1
        return unquoted

    def _fetchOne(self, sql):

        return self.core_connection.executeSql(sql)

    def _fetchAll(self, sql, include_fid_and_geometry=False):

        return self.core_connection.executeSql(sql)

    def _fetchAllFromLayer(self, table):

        lyr = self.gdal_ds.GetLayerByName(table.name)
        if lyr is None:
            return []

        lyr.ResetReading()
        ret = []
        while True:
            f = lyr.GetNextFeature()
            if f is None:
                break
            else:
                geom_name = lyr.GetGeometryColumn()
                fid_name = lyr.GetFIDColumn()

                field_vals = []
                field_names = [x.name for x in table.fields()]
                for field_name in field_names:
                    if fid_name and fid_name == field_name:
                        field_vals.append(f.GetFID())
                    elif geom_name and geom_name == field_name:
                        geom = f.GetGeometryRef()
                        field_vals.append(geom.ExportToWkt())
                    else:
                        field_index = f.GetFieldIndex(field_name)
                        field_vals.append(f.GetField(field_index))
                ret.append(field_vals)
        return ret

    def _execute_and_commit(self, sql):
        sql_lyr = self.gdal_ds.ExecuteSQL(sql)
        self.gdal_ds.ReleaseResultSet(sql_lyr)

    def _execute(self, cursor, sql):

        if self.connection is None:
            # Needed when evaluating a SQL query
            try:
                self.connection = spatialite_connect(str(self.dbname))
            except self.connection_error_types() as e:
                raise ConnectionError(e)

        return DBConnector._execute(self, cursor, sql)

    def _commit(self):
        if self.connection is None:
            return

        try:
            self.connection.commit()

        except self.connection_error_types() as e:
            raise ConnectionError(e)

        except self.execution_error_types() as e:
            # do the rollback to avoid a "current transaction aborted, commands ignored" errors
            self._rollback()
            raise DbError(e)

    def cancel(self):
        if self.connection:
            self.connection.interrupt()

    @classmethod
    def isValidDatabase(cls, path):
        try:
            ds = gdal.OpenEx(path)
        except Exception:
            return False
        return ds.GetDriver().ShortName == "GPKG"

    def getInfo(self):
        return None

    def getSpatialInfo(self):
        return None

    def hasSpatialSupport(self):
        return True

    # Used by DlgTableProperties
    def canAddGeometryColumn(self, table):
        _, tablename = self.getSchemaTableName(table)
        lyr = self.gdal_ds.GetLayerByName(tablename)
        if lyr is None:
            return False
        return lyr.GetGeomType() == ogr.wkbNone

    # Used by DlgTableProperties
    def canAddSpatialIndex(self, table):
        _, tablename = self.getSchemaTableName(table)
        lyr = self.gdal_ds.GetLayerByName(tablename)
        if lyr is None or lyr.GetGeometryColumn() == "":
            return False
        return not self.hasSpatialIndex(table, lyr.GetGeometryColumn())

    def hasRasterSupport(self):
        return self.has_raster

    def hasCustomQuerySupport(self):
        return True

    def hasTableColumnEditingSupport(self):
        return True

    def hasCreateSpatialViewSupport(self):
        return False

    def fieldTypes(self):
        # From "Table 1. GeoPackage Data Types" (http://www.geopackage.org/spec/)
        return [
            "TEXT",
            "MEDIUMINT",
            "INTEGER",
            "TINYINT",
            "SMALLINT",
            "DOUBLE",
            "FLOATDATE",
            "DATETIME",
            "BOOLEAN",
        ]

    def getSchemas(self):
        return None

    def getTables(self, schema=None, add_sys_tables=False):
        """get list of tables"""
        items = []

        try:
            vectors = self.getVectorTables(schema)
            for tbl in vectors:
                items.append(tbl)
        except DbError:
            pass

        try:
            rasters = self.getRasterTables(schema)
            for tbl in rasters:
                items.append(tbl)
        except DbError:
            pass

        for i, tbl in enumerate(items):
            tbl.insert(3, False)  # not system table

        return sorted(items, key=cmp_to_key(lambda x, y: (x[1] > y[1]) - (x[1] < y[1])))

    def getVectorTables(self, schema=None):
        """Returns a list of vector table information"""

        items = []
        for table in self.core_connection.tables(
            schema,
            QgsAbstractDatabaseProviderConnection.TableFlag.Vector
            | QgsAbstractDatabaseProviderConnection.TableFlag.Aspatial,
        ):
            if not (
                table.flags() & QgsAbstractDatabaseProviderConnection.TableFlag.Aspatial
            ):
                geom_type = table.geometryColumnTypes()[0]
                # Use integer PG code for SRID
                srid = geom_type.crs.postgisSrid()
                geomtype_flatten = QgsWkbTypes.flatType(geom_type.wkbType)
                geomname = "GEOMETRY"
                if geomtype_flatten == QgsWkbTypes.Type.Point:
                    geomname = "POINT"
                elif geomtype_flatten == QgsWkbTypes.Type.LineString:
                    geomname = "LINESTRING"
                elif geomtype_flatten == QgsWkbTypes.Type.Polygon:
                    geomname = "POLYGON"
                elif geomtype_flatten == QgsWkbTypes.Type.MultiPoint:
                    geomname = "MULTIPOINT"
                elif geomtype_flatten == QgsWkbTypes.Type.MultiLineString:
                    geomname = "MULTILINESTRING"
                elif geomtype_flatten == QgsWkbTypes.Type.MultiPolygon:
                    geomname = "MULTIPOLYGON"
                elif geomtype_flatten == QgsWkbTypes.Type.GeometryCollection:
                    geomname = "GEOMETRYCOLLECTION"
                elif geomtype_flatten == QgsWkbTypes.Type.CircularString:
                    geomname = "CIRCULARSTRING"
                elif geomtype_flatten == QgsWkbTypes.Type.CompoundCurve:
                    geomname = "COMPOUNDCURVE"
                elif geomtype_flatten == QgsWkbTypes.Type.CurvePolygon:
                    geomname = "CURVEPOLYGON"
                elif geomtype_flatten == QgsWkbTypes.Type.MultiCurve:
                    geomname = "MULTICURVE"
                elif geomtype_flatten == QgsWkbTypes.Type.MultiSurface:
                    geomname = "MULTISURFACE"
                geomdim = "XY"
                if QgsWkbTypes.hasZ(geom_type.wkbType):
                    geomdim += "Z"
                if QgsWkbTypes.hasM(geom_type.wkbType):
                    geomdim += "M"
                item = [
                    Table.VectorType,
                    table.tableName(),
                    bool(
                        table.flags()
                        & QgsAbstractDatabaseProviderConnection.TableFlag.View
                    ),  # is_view
                    table.tableName(),
                    table.geometryColumn(),
                    geomname,
                    geomdim,
                    srid,
                ]
                self.mapSridToName[srid] = geom_type.crs.description()
            else:
                item = [
                    Table.TableType,
                    table.tableName(),
                    bool(
                        table.flags()
                        & QgsAbstractDatabaseProviderConnection.TableFlag.View
                    ),
                ]

            items.append(item)

        return items

    def getRasterTables(self, schema=None):
        """get list of table with a geometry column
        it returns:
                name (table name)
                type = 'view' (is a view?)
                geometry_column:
                        r.table_name (the prefix table name, use this to load the layer)
                        r.geometry_column
                        srid
        """

        items = []
        for table in self.core_connection.tables(
            schema, QgsAbstractDatabaseProviderConnection.TableFlag.Raster
        ):
            geom_type = table.geometryColumnTypes()[0]
            # Use integer PG code for SRID
            srid = geom_type.crs.postgisSrid()
            item = [
                Table.RasterType,
                table.tableName(),
                bool(
                    table.flags() & QgsAbstractDatabaseProviderConnection.TableFlag.View
                ),
                table.tableName(),
                table.geometryColumn(),
                srid,
            ]
            self.mapSridToName[srid] = geom_type.crs.description()
            items.append(item)

        return items

    def getTableRowCount(self, table):
        lyr = self.gdal_ds.GetLayerByName(self.getSchemaTableName(table)[1])
        return lyr.GetFeatureCount() if lyr is not None else None

    def getTableFields(self, table):
        """return list of columns in table"""
        sql = "PRAGMA table_info(%s)" % (self.quoteId(table))
        ret = self._fetchAll(sql)
        if ret is None:
            ret = []
        return ret

    def getTableIndexes(self, table):
        """get info about table's indexes"""
        sql = "PRAGMA index_list(%s)" % (self.quoteId(table))
        indexes = self._fetchAll(sql)
        if indexes is None:
            return []

        for i, idx in enumerate(indexes):
            # sqlite has changed the number of columns returned by index_list since 3.8.9
            # I am not using self.getInfo() here because this behavior
            # can be changed back without notice as done for index_info, see:
            # http://repo.or.cz/sqlite.git/commit/53555d6da78e52a430b1884b5971fef33e9ccca4
            if len(idx) == 3:
                num, name, unique = idx
            if len(idx) == 5:
                num, name, unique, createdby, partial = idx
            sql = "PRAGMA index_info(%s)" % (self.quoteId(name))

            idx = [num, name, unique]
            cols = [cid for seq, cid, cname in self._fetchAll(sql)]
            idx.append(cols)
            indexes[i] = idx

        return indexes

    def getTableConstraints(self, table):
        return None

    def getTableTriggers(self, table):

        _, tablename = self.getSchemaTableName(table)
        # Do not list rtree related triggers as we don't want them to be dropped
        sql = (
            "SELECT name, sql FROM sqlite_master WHERE tbl_name = %s AND type = 'trigger'"
            % (self.quoteString(tablename))
        )
        if self.isVectorTable(table):
            sql += " AND name NOT LIKE 'rtree_%%'"
        elif self.isRasterTable(table):
            sql += " AND name NOT LIKE '%%_zoom_insert'"
            sql += " AND name NOT LIKE '%%_zoom_update'"
            sql += " AND name NOT LIKE '%%_tile_column_insert'"
            sql += " AND name NOT LIKE '%%_tile_column_update'"
            sql += " AND name NOT LIKE '%%_tile_row_insert'"
            sql += " AND name NOT LIKE '%%_tile_row_update'"
        return self._fetchAll(sql)

    def deleteTableTrigger(self, trigger, table=None):
        """Deletes trigger"""
        sql = "DROP TRIGGER %s" % self.quoteId(trigger)
        self._execute_and_commit(sql)

    def getTableExtent(self, table, geom, force=False):
        """find out table extent"""
        _, tablename = self.getSchemaTableName(table)

        if self.isRasterTable(table):
            md = self.gdal_ds.GetMetadata("SUBDATASETS")
            if md is None or len(md) == 0:
                ds = self.gdal_ds
            else:
                subdataset_name = f"GPKG:{self.gdal_ds.GetDescription()}:{tablename}"
                try:
                    ds = gdal.Open(subdataset_name)
                except Exception:
                    ds = None
            if ds is None:
                return None
            gt = ds.GetGeoTransform()
            minx = gt[0]
            maxx = gt[0] + gt[1] * ds.RasterYSize
            maxy = gt[3]
            miny = gt[3] + gt[5] * ds.RasterYSize

            return (minx, miny, maxx, maxy)

        lyr = self.gdal_ds.GetLayerByName(tablename)
        if lyr is None:
            return None
        ret = lyr.GetExtent(force=force, can_return_null=True)
        if ret is None:
            return None
        minx, maxx, miny, maxy = ret
        return (minx, miny, maxx, maxy)

    def getViewDefinition(self, view):
        """returns definition of the view"""
        return None

    def getSpatialRefInfo(self, srid):
        if srid in self.mapSridToName:
            return self.mapSridToName[srid]

        sql = (
            "SELECT srs_name FROM gpkg_spatial_ref_sys WHERE srs_id = %s"
            % self.quoteString(srid)
        )
        res = self._fetchOne(sql)
        if res is not None and len(res) > 0:
            res = res[0]
        self.mapSridToName[srid] = res
        return res

    def isVectorTable(self, table):

        _, tablename = self.getSchemaTableName(table)
        return self.gdal_ds.GetLayerByName(tablename) is not None

    def isRasterTable(self, table):
        if self.has_raster and not self.isVectorTable(table):
            _, tablename = self.getSchemaTableName(table)
            md = self.gdal_ds.GetMetadata("SUBDATASETS")
            if md is None or len(md) == 0:
                sql = (
                    "SELECT COUNT(*) FROM gpkg_contents WHERE data_type = 'tiles' AND table_name = %s"
                    % self.quoteString(tablename)
                )
                ret = self._fetchOne(sql)
                return ret != [] and ret[0][0] == 1
            else:
                subdataset_name = f"GPKG:{self.gdal_ds.GetDescription()}:{tablename}"
                for key in md:
                    if md[key] == subdataset_name:
                        return True

        return False

    def getOGRFieldTypeFromSQL(self, sql_type):
        ogr_type = ogr.OFTString
        ogr_subtype = ogr.OFSTNone
        width = 0
        if not sql_type.startswith("TEXT ("):
            pos = sql_type.find(" (")
            if pos >= 0:
                sql_type = sql_type[0:pos]
        if sql_type == "BOOLEAN":
            ogr_type = ogr.OFTInteger
            ogr_subtype = ogr.OFSTBoolean
        elif sql_type in ("TINYINT", "SMALLINT", "MEDIUMINT"):
            ogr_type = ogr.OFTInteger
        elif sql_type == "INTEGER":
            ogr_type = ogr.OFTInteger64
        elif sql_type == "FLOAT":
            ogr_type = ogr.OFTReal
            ogr_subtype = ogr.OFSTFloat32
        elif sql_type == "DOUBLE":
            ogr_type = ogr.OFTReal
        elif sql_type == "DATE":
            ogr_type = ogr.OFTDate
        elif sql_type == "DATETIME":
            ogr_type = ogr.OFTDateTime
        elif sql_type.startswith("TEXT (") and sql_type.endswith(")"):
            width = int(sql_type[len("TEXT (") : -1])
        return (ogr_type, ogr_subtype, width)

    def createOGRFieldDefnFromSQL(self, sql_fielddef):
        f_split = sql_fielddef.split(" ")
        quoted_name = f_split[0]
        name = self.unquoteId(quoted_name)
        sql_type = f_split[1].upper()
        if (
            len(f_split) >= 3
            and f_split[2].startswith("(")
            and f_split[2].endswith(")")
        ):
            sql_type += " " + f_split[2]
            f_split = [f for f in f_split[3:]]
        else:
            f_split = [f for f in f_split[2:]]
        ogr_type, ogr_subtype, width = self.getOGRFieldTypeFromSQL(sql_type)
        fld_defn = ogr.FieldDefn(name, ogr_type)
        fld_defn.SetSubType(ogr_subtype)
        fld_defn.SetWidth(width)
        if len(f_split) >= 2 and f_split[0] == "NOT" and f_split[1] == "NULL":
            fld_defn.SetNullable(False)
            f_split = [f for f in f_split[2:]]
        elif len(f_split) >= 1:
            f_split = [f for f in f_split[1:]]
        if len(f_split) >= 2 and f_split[0] == "DEFAULT":
            new_default = f_split[1]
            if new_default == "":
                fld_defn.SetDefault(None)
            elif new_default == "NULL" or ogr_type in (ogr.OFTInteger, ogr.OFTReal):
                fld_defn.SetDefault(new_default)
            elif new_default.startswith("'") and new_default.endswith("'"):
                fld_defn.SetDefault(new_default)
            else:
                fld_defn.SetDefault(self.quoteString(new_default))
        return fld_defn

    def createTable(self, table, field_defs, pkey):
        """Creates ordinary table
        'fields' is array containing field definitions
        'pkey' is the primary key name
        """
        if len(field_defs) == 0:
            return False

        options = []
        if pkey is not None and pkey != "":
            options += ["FID=" + pkey]
        _, tablename = self.getSchemaTableName(table)
        lyr = self.gdal_ds.CreateLayer(
            tablename, geom_type=ogr.wkbNone, options=options
        )
        if lyr is None:
            return False
        for field_def in field_defs:
            fld_defn = self.createOGRFieldDefnFromSQL(field_def)
            if fld_defn.GetName() == pkey:
                continue
            if lyr.CreateField(fld_defn) != 0:
                return False

        return True

    def deleteTable(self, table):
        """Deletes table from the database"""
        if self.isRasterTable(table):
            sql = f"DROP TABLE {self.quoteId(table)}"
            self._execute_and_commit(sql)
            return True

        _, tablename = self.getSchemaTableName(table)
        for i in range(self.gdal_ds.GetLayerCount()):
            if self.gdal_ds.GetLayer(i).GetName() == tablename:
                return self.gdal_ds.DeleteLayer(i) == 0
        return False

    def emptyTable(self, table):
        """Deletes all rows from table"""
        if self.isRasterTable(table):
            return False

        sql = "DELETE FROM %s" % self.quoteId(table)
        self._execute_and_commit(sql)

    def renameTable(self, table, new_table):
        """Renames the table

        :param table: tuple with schema and table names
        :type table: tuple (str, str)
        :param new_table: new table name
        :type new_table: str
        :return: true on success
        :rtype: bool
        """
        try:
            name = table[1]  # 0 is schema
            vector_table_names = [
                t.tableName()
                for t in self.core_connection.tables(
                    "", QgsAbstractDatabaseProviderConnection.TableFlag.Vector
                )
            ]
            if name in vector_table_names:
                self.core_connection.renameVectorTable("", name, new_table)
            else:
                self.core_connection.renameRasterTable("", name, new_table)
            return True
        except QgsProviderConnectionException:
            return False

    def moveTable(self, table, new_table, new_schema=None):
        return self.renameTable(table, new_table)

    def runVacuum(self):
        """run vacuum on the db"""
        self._execute_and_commit("VACUUM")

    def addTableColumn(self, table, field_def):
        """Adds a column to table"""

        _, tablename = self.getSchemaTableName(table)
        lyr = self.gdal_ds.GetLayerByName(tablename)
        if lyr is None:
            return False
        fld_defn = self.createOGRFieldDefnFromSQL(field_def)
        return lyr.CreateField(fld_defn) == 0

    def deleteTableColumn(self, table, column):
        """Deletes column from a table"""
        if self.isGeometryColumn(table, column):
            return False

        _, tablename = self.getSchemaTableName(table)
        lyr = self.gdal_ds.GetLayerByName(tablename)
        if lyr is None:
            return False
        idx = lyr.GetLayerDefn().GetFieldIndex(column)
        if idx >= 0:
            return lyr.DeleteField(idx) == 0
        return False

    def updateTableColumn(
        self,
        table,
        column,
        new_name,
        new_data_type=None,
        new_not_null=None,
        new_default=None,
        comment=None,
    ):
        if self.isGeometryColumn(table, column):
            return False

        _, tablename = self.getSchemaTableName(table)
        lyr = self.gdal_ds.GetLayerByName(tablename)
        if lyr is None:
            return False
        if lyr.TestCapability(ogr.OLCAlterFieldDefn) == 0:
            return False
        idx = lyr.GetLayerDefn().GetFieldIndex(column)
        if idx >= 0:
            old_fielddefn = lyr.GetLayerDefn().GetFieldDefn(idx)
            flag = 0
            if new_name is not None:
                flag |= ogr.ALTER_NAME_FLAG
            else:
                new_name = column
            if new_data_type is None:
                ogr_type = old_fielddefn.GetType()
                ogr_subtype = old_fielddefn.GetSubType()
                width = old_fielddefn.GetWidth()
            else:
                flag |= ogr.ALTER_TYPE_FLAG
                flag |= ogr.ALTER_WIDTH_PRECISION_FLAG
                ogr_type, ogr_subtype, width = self.getOGRFieldTypeFromSQL(
                    new_data_type
                )
            new_fielddefn = ogr.FieldDefn(new_name, ogr_type)
            new_fielddefn.SetSubType(ogr_subtype)
            new_fielddefn.SetWidth(width)
            if new_default is not None:
                flag |= ogr.ALTER_DEFAULT_FLAG
                if new_default == "":
                    new_fielddefn.SetDefault(None)
                elif new_default == "NULL" or ogr_type in (ogr.OFTInteger, ogr.OFTReal):
                    new_fielddefn.SetDefault(str(new_default))
                elif new_default.startswith("'") and new_default.endswith("'"):
                    new_fielddefn.SetDefault(str(new_default))
                else:
                    new_fielddefn.SetDefault(self.quoteString(new_default))
            else:
                new_fielddefn.SetDefault(old_fielddefn.GetDefault())
            if new_not_null is not None:
                flag |= ogr.ALTER_NULLABLE_FLAG
                new_fielddefn.SetNullable(not new_not_null)
            else:
                new_fielddefn.SetNullable(old_fielddefn.IsNullable())
            return lyr.AlterFieldDefn(idx, new_fielddefn, flag) == 0

        return False

    def isGeometryColumn(self, table, column):

        _, tablename = self.getSchemaTableName(table)
        lyr = self.gdal_ds.GetLayerByName(tablename)
        if lyr is None:
            return False
        return column == lyr.GetGeometryColumn()

    def addGeometryColumn(
        self, table, geom_column="geometry", geom_type="POINT", srid=-1, dim=2
    ):

        _, tablename = self.getSchemaTableName(table)
        lyr = self.gdal_ds.GetLayerByName(tablename)
        if lyr is None:
            return False
        ogr_type = ogr.wkbUnknown
        if geom_type == "POINT":
            ogr_type = ogr.wkbPoint
        elif geom_type == "LINESTRING":
            ogr_type = ogr.wkbLineString
        elif geom_type == "POLYGON":
            ogr_type = ogr.wkbPolygon
        elif geom_type == "MULTIPOINT":
            ogr_type = ogr.wkbMultiPoint
        elif geom_type == "MULTILINESTRING":
            ogr_type = ogr.wkbMultiLineString
        elif geom_type == "MULTIPOLYGON":
            ogr_type = ogr.wkbMultiPolygon
        elif geom_type == "GEOMETRYCOLLECTION":
            ogr_type = ogr.wkbGeometryCollection

        if dim == 3:
            ogr_type = ogr_type | ogr.wkb25DBit
        elif dim == 4:
            if hasattr(ogr, "GT_HasZ"):
                ogr_type = ogr.GT_SetZ(ogr_type)
            else:
                ogr_type = ogr_type | ogr.wkb25DBit
            if hasattr(ogr, "GT_HasM"):
                ogr_type = ogr.GT_SetM(ogr_type)

        geom_field_defn = ogr.GeomFieldDefn(self.unquoteId(geom_column), ogr_type)
        if srid > 0:
            sr = osr.SpatialReference()
            if sr.ImportFromEPSG(srid) == 0:
                geom_field_defn.SetSpatialRef(sr)

        if lyr.CreateGeomField(geom_field_defn) != 0:
            return False
        self._opendb()
        return True

    def deleteGeometryColumn(self, table, geom_column):
        return False  # not supported

    def addTableUniqueConstraint(self, table, column):
        """Adds a unique constraint to a table"""
        return False  # constraints not supported

    def deleteTableConstraint(self, table, constraint):
        """Deletes constraint in a table"""
        return False  # constraints not supported

    def addTablePrimaryKey(self, table, column):
        """Adds a primery key (with one column) to a table"""
        sql = f"ALTER TABLE {self.quoteId(table)} ADD PRIMARY KEY ({self.quoteId(column)})"
        self._execute_and_commit(sql)

    def createTableIndex(self, table, name, column, unique=False):
        """Creates index on one column using default options"""
        unique_str = "UNIQUE" if unique else ""
        sql = f"CREATE {unique_str} INDEX {self.quoteId(name)} ON {self.quoteId(table)} ({self.quoteId(column)})"
        self._execute_and_commit(sql)

    def deleteTableIndex(self, table, name):
        schema, tablename = self.getSchemaTableName(table)
        sql = "DROP INDEX %s" % self.quoteId((schema, name))
        self._execute_and_commit(sql)

    def createSpatialIndex(self, table, geom_column):
        if self.isRasterTable(table):
            return False
        _, tablename = self.getSchemaTableName(table)
        sql = f"SELECT CreateSpatialIndex({self.quoteId(tablename)}, {self.quoteId(geom_column)})"
        try:
            res = self._fetchOne(sql)
        except QgsProviderConnectionException:
            return False
        return res is not None and res[0][0] == 1

    def deleteSpatialIndex(self, table, geom_column):
        if self.isRasterTable(table):
            return False
        _, tablename = self.getSchemaTableName(table)
        sql = f"SELECT DisableSpatialIndex({self.quoteId(tablename)}, {self.quoteId(geom_column)})"
        res = self._fetchOne(sql)
        return len(res) > 0 and len(res[0]) > 0 and res[0][0] == 1

    def hasSpatialIndex(self, table, geom_column):
        if self.isRasterTable(table) or geom_column is None:
            return False
        _, tablename = self.getSchemaTableName(table)

        # (only available in >= 2.1.2)
        sql = f"SELECT HasSpatialIndex({self.quoteString(tablename)}, {self.quoteString(geom_column)})"
        gdal.PushErrorHandler()
        ret = self._fetchOne(sql)
        gdal.PopErrorHandler()

        if len(ret) == 0:
            # might be the case for GDAL < 2.1.2
            sql = (
                "SELECT COUNT(*) FROM sqlite_master WHERE type = 'table' AND name LIKE %s"
                % self.quoteString("%%rtree_" + tablename + "_%%")
            )
            ret = self._fetchOne(sql)
        if len(ret) == 0:
            return False
        else:
            return ret[0][0] >= 1

    def execution_error_types(self):
        return sqlite3.Error, sqlite3.ProgrammingError, sqlite3.Warning

    def connection_error_types(self):
        return sqlite3.InterfaceError, sqlite3.OperationalError

    def getSqlDictionary(self):
        from .sql_dictionary import getSqlDictionary

        sql_dict = getSqlDictionary()

        items = []
        for tbl in self.getTables():
            items.append(tbl[1])  # table name

            for fld in self.getTableFields(tbl[0]):
                items.append(fld[1])  # field name

        sql_dict["identifier"] = items
        return sql_dict

    def getQueryBuilderDictionary(self):
        from .sql_dictionary import getQueryBuilderDictionary

        return getQueryBuilderDictionary()



"""
***************************************************************************
    gdal2xyz.py
    ---------------------
    Date                 : September 2013
    Copyright            : (C) 2013 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "September 2013"
__copyright__ = "(C) 2013, Alexander Bruy"

from qgis.core import (
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingParameterBand,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterFileDestination,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterLayer,
)

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils
from processing.tools.system import isWindows


class gdal2xyz(GdalAlgorithm):
    INPUT = "INPUT"
    BAND = "BAND"
    SRCNODATA = "NODATA_INPUT"
    DSTNODATA = "NODATA_OUTPUT"
    SKIPNODATA = "SKIP_NODATA"
    CSV = "CSV"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterBand(
                self.BAND,
                self.tr("Band number"),
                1,
                parentLayerParameterName=self.INPUT,
            )
        )

        self.addParameter(
            QgsProcessingParameterNumber(
                self.SRCNODATA,
                self.tr("Input pixel value to treat as NoData"),
                optional=True,
            )
        )  # GDAL > 3.6.3
        self.addParameter(
            QgsProcessingParameterNumber(
                self.DSTNODATA,
                self.tr("Assign specified NoData value to output"),
                optional=True,
            )
        )  # GDAL > 3.6.3
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.SKIPNODATA,
                self.tr("Do not output NoData values"),
                defaultValue=False,
            )
        )  # GDAL > 3.3
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.CSV, self.tr("Output comma-separated values"), defaultValue=False
            )
        )
        self.addParameter(
            QgsProcessingParameterFileDestination(
                self.OUTPUT, self.tr("XYZ ASCII file"), self.tr("CSV files (*.csv)")
            )
        )

    def name(self):
        return "gdal2xyz"

    def displayName(self):
        return self.tr("gdal2xyz")

    def group(self):
        return self.tr("Raster conversion")

    def groupId(self):
        return "rasterconversion"

    def commandName(self):
        return "gdal2xyz"

    def flags(self):
        return super().flags() | QgsProcessingAlgorithm.Flag.FlagDisplayNameIsLiteral

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        arguments = ["-band", str(self.parameterAsInt(parameters, self.BAND, context))]

        if self.SRCNODATA in parameters and parameters[self.SRCNODATA] is not None:
            if (
                GdalUtils.version() > 3060300
            ):  # src/dstnodata broken <= 3.6.3 https://github.com/OSGeo/gdal/issues/7410
                srcnodata = self.parameterAsDouble(parameters, self.SRCNODATA, context)
                arguments.append("-srcnodata")
                arguments.append(srcnodata)
            else:
                raise QgsProcessingException(
                    self.tr(
                        "The source nodata option (-srcnodata) is only available on GDAL 3.6.4 or later"
                    )
                )

        if self.DSTNODATA in parameters and parameters[self.DSTNODATA] is not None:
            if (
                GdalUtils.version() > 3060300
            ):  # src/dstnodata broken <= 3.6.3 https://github.com/OSGeo/gdal/issues/7410
                dstnodata = self.parameterAsDouble(parameters, self.DSTNODATA, context)
                arguments.append("-dstnodata")
                arguments.append(dstnodata)
            else:
                raise QgsProcessingException(
                    self.tr(
                        "The destination nodata option (-dstnodata) is only available on GDAL 3.6.4 or later"
                    )
                )

        if self.SKIPNODATA in parameters:
            if GdalUtils.version() >= 3030000:  # skipnodata added at GDAL 3.3
                if self.parameterAsBoolean(parameters, self.SKIPNODATA, context):
                    arguments.append("-skipnodata")
            else:
                raise QgsProcessingException(
                    self.tr(
                        "The skip nodata option (-skipnodata) is only available on GDAL 3.3 or later"
                    )
                )

        if self.parameterAsBoolean(parameters, self.CSV, context):
            arguments.append("-csv")

        raster = self.parameterAsRasterLayer(parameters, self.INPUT, context)
        if raster is None:
            raise QgsProcessingException(
                self.invalidRasterError(parameters, self.INPUT)
            )

        input_details = GdalUtils.gdal_connection_details_from_layer(raster)

        arguments.append(input_details.connection_string)
        arguments.append(self.parameterAsFileOutput(parameters, self.OUTPUT, context))

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        return [
            self.commandName() + (".bat" if isWindows() else ".py"),
            GdalUtils.escapeAndJoin(arguments),
        ]



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : Oct 13, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

The content of this file is based on
- PG_Manager by Martin Dobias (GPLv2 license)
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.core import (
    QgsCoordinateReferenceSystem,
    QgsSettings,
    QgsVectorDataProvider,
    QgsVectorFileWriter,
    QgsVectorLayerExporter,
)
from qgis.PyQt import uic
from qgis.PyQt.QtCore import QFileInfo, Qt
from qgis.PyQt.QtGui import QCursor
from qgis.PyQt.QtWidgets import QApplication, QDialog, QFileDialog, QMessageBox
from qgis.utils import OverrideCursor

from .gui_utils import GuiUtils

Ui_Dialog, _ = uic.loadUiType(GuiUtils.get_ui_file_path("DlgExportVector.ui"))


class DlgExportVector(QDialog, Ui_Dialog):
    def __init__(self, inLayer, inDb, parent=None):
        QDialog.__init__(self, parent)
        self.inLayer = inLayer
        self.db = inDb
        self.setupUi(self)

        vectorFilterName = "lastVectorFileFilter"  # "lastRasterFileFilter"
        self.lastUsedVectorFilterSettingsKey = f"/UI/{vectorFilterName}"
        self.lastUsedVectorDirSettingsKey = f"/UI/{vectorFilterName}Dir"

        # update UI
        self.setupWorkingMode()
        self.populateFileFilters()
        self.populateEncodings()

    def setupWorkingMode(self):
        # set default values
        inCrs = self.inLayer.crs()
        srid = inCrs.postgisSrid() if inCrs.isValid() else 4236
        self.editSourceSrid.setText("%s" % srid)
        self.editTargetSrid.setText("%s" % srid)

        self.btnChooseOutputFile.clicked.connect(self.chooseOutputFile)
        self.checkSupports()

    def checkSupports(self):
        """update options available for the current input layer"""
        allowSpatial = self.db.connector.hasSpatialSupport()
        hasGeomType = self.inLayer and self.inLayer.isSpatial()
        self.chkSourceSrid.setEnabled(allowSpatial and hasGeomType)
        self.chkTargetSrid.setEnabled(allowSpatial and hasGeomType)
        # self.chkSpatialIndex.setEnabled(allowSpatial and hasGeomType)

    def chooseOutputFile(self):
        # get last used dir
        settings = QgsSettings()
        lastUsedDir = settings.value(self.lastUsedVectorDirSettingsKey, ".")

        # get selected filter
        selected_driver = self.cboFileFormat.currentData()
        selected_filter = QgsVectorFileWriter.filterForDriver(selected_driver)

        # ask for a filename
        filename, filter = QFileDialog.getSaveFileName(
            self, self.tr("Choose where to save the file"), lastUsedDir, selected_filter
        )
        if filename == "":
            return

        ext = selected_filter[selected_filter.find(".") :]
        ext = ext[: ext.find(" ")]

        if not filename.lower().endswith(ext):
            filename += ext

        # store the last used dir
        settings.setValue(
            self.lastUsedVectorDirSettingsKey, QFileInfo(filename).filePath()
        )

        self.editOutputFile.setText(filename)

    def populateEncodings(self):
        # populate the combo with supported encodings
        self.cboEncoding.addItems(QgsVectorDataProvider.availableEncodings())

        # set the last used encoding
        enc = self.inLayer.dataProvider().encoding()
        idx = self.cboEncoding.findText(enc)
        if idx < 0:
            self.cboEncoding.insertItem(0, enc)
            idx = 0
        self.cboEncoding.setCurrentIndex(idx)

    def populateFileFilters(self):
        # populate the combo with supported vector file formats
        for driver in QgsVectorFileWriter.ogrDriverList():
            self.cboFileFormat.addItem(driver.longName, driver.driverName)

        # set the last used filter
        settings = QgsSettings()
        filt = settings.value(self.lastUsedVectorFilterSettingsKey, "GPKG")

        idx = self.cboFileFormat.findText(filt)
        if idx < 0:
            idx = 0
        self.cboFileFormat.setCurrentIndex(idx)

    def accept(self):
        # sanity checks
        if self.editOutputFile.text() == "":
            QMessageBox.information(
                self, self.tr("Export to file"), self.tr("Output file name is required")
            )
            return

        if self.chkSourceSrid.isEnabled() and self.chkSourceSrid.isChecked():
            try:
                sourceSrid = int(self.editSourceSrid.text())
            except ValueError:
                QMessageBox.information(
                    self,
                    self.tr("Export to file"),
                    self.tr("Invalid source srid: must be an integer"),
                )
                return

        if self.chkTargetSrid.isEnabled() and self.chkTargetSrid.isChecked():
            try:
                targetSrid = int(self.editTargetSrid.text())
            except ValueError:
                QMessageBox.information(
                    self,
                    self.tr("Export to file"),
                    self.tr("Invalid target srid: must be an integer"),
                )
                return

        with OverrideCursor(Qt.CursorShape.WaitCursor):
            # store current input layer crs, so I can restore it later
            prevInCrs = self.inLayer.crs()
            try:
                uri = self.editOutputFile.text()
                providerName = "ogr"

                options = {}

                # set the OGR driver will be used
                driverName = self.cboFileFormat.currentData()
                options["driverName"] = driverName

                # set the output file encoding
                if self.chkEncoding.isEnabled() and self.chkEncoding.isChecked():
                    enc = self.cboEncoding.currentText()
                    options["fileEncoding"] = enc

                if self.chkDropTable.isChecked():
                    options["overwrite"] = True

                outCrs = QgsCoordinateReferenceSystem()
                if self.chkTargetSrid.isEnabled() and self.chkTargetSrid.isChecked():
                    targetSrid = int(self.editTargetSrid.text())
                    outCrs = QgsCoordinateReferenceSystem(targetSrid)

                # update input layer crs
                if self.chkSourceSrid.isEnabled() and self.chkSourceSrid.isChecked():
                    sourceSrid = int(self.editSourceSrid.text())
                    inCrs = QgsCoordinateReferenceSystem(sourceSrid)
                    self.inLayer.setCrs(inCrs)

                # do the export!
                ret, errMsg = QgsVectorLayerExporter.exportLayer(
                    self.inLayer, uri, providerName, outCrs, False, options
                )
            except Exception as e:
                ret = -1
                errMsg = str(e)

            finally:
                # restore input layer crs and encoding
                self.inLayer.setCrs(prevInCrs)

        if ret != 0:
            QMessageBox.warning(
                self,
                self.tr("Export to file"),
                self.tr("Error {0}\n{1}").format(ret, errMsg),
            )
            return

        # create spatial index
        # if self.chkSpatialIndex.isEnabled() and self.chkSpatialIndex.isChecked():
        #       self.db.connector.createSpatialIndex( (schema, table), geom )

        QMessageBox.information(
            self, self.tr("Export to file"), self.tr("Export finished.")
        )
        return QDialog.accept(self)
